/*--[drm5.c]-------------------------------------------------------------------
 | Copyright (C) 2002, 2003 Dan A. Jackson 
 |
 | This file is part of the "clit" (Convert LIT) program. 
 |
 | Convert LIT is free software; you can redistribute it and/or modify
 | it under the terms of the GNU General Public License as published by
 | the Free Software Foundation; either version 2 of the License, or
 | (at your option) any later version.
 |
 | This program is distributed in the hope that it will be useful,
 | but WITHOUT ANY WARRANTY; without even the implied warranty of
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 | GNU General Public License for more details.
 |
 | You should have received a copy of the GNU General Public License
 | along with this program; if not, write to the Free Software
 | Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 | 
 | The GNU General Public License may also be available at the following
 | URL: http://www.gnu.org/licenses/gpl.html
*/
/* This file contains a way to slip through the DRM5 fences.  It's only 
 | a stop-gap measure, and there are always other ways.   */

/* 
 | Update for June 4th, the day Americans celebrate their freedom to write
 | software that restores fair-use rights.   
 | 
 | After a 6 month hiatus Microsoft released an update which checked that
 | the filename calling the Secure Repository DLL was actually inside the
 | manifest.  Unfortunately, ConvertLit is a terrible lying cheat of a
 | program, and tells the DLL exactly what it wants to hear.
 */
/*
 * Now uses an external DLL.  This should improve stability and resilience,
 * at the cost of having to worry about 2 files. 
 */


#ifdef _MSC_VER
#include <windows.h>
#include <winbase.h>
#include <winnt.h>      /* PE-Format defines */
#include <direct.h>
#endif

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "litlib.h"

static const char * license_string = "/DRMStorage/Licenses/EUL";


typedef int (*p_drm5hook_decrypt)(unsigned char *, int len);
typedef int (*p_drm5hook_init)(void);
typedef int (*p_drm5hook_install)(char *, void *);
typedef int (*p_drm5hook_setkey)(unsigned char *, int len, DWORD extra);

p_drm5hook_decrypt drm5hook_decrypt;
p_drm5hook_init drm5hook_init;
p_drm5hook_install drm5hook_install;
p_drm5hook_setkey drm5hook_setkey;

int MS_Base64Decode(wchar_t * str, unsigned char **buff)
{
    wchar_t *cp;
    char *ocp;
    int len, val, count, block, ocount;

    len = wcslen(str);
    if ((*buff = malloc((len * 3) / 4)) == NULL) return -1;
    ocp = *buff;
    count = 0;
    block = 0;
    ocount = 0;
    for (cp = str; *cp != L'\0'; cp++) {
        if ((*cp >= L'A') && (*cp <= L'Z'))
            val = *cp - L'A';
        else if ((*cp >= L'a') && (*cp <= L'z'))
            val = *cp - L'a' + 26;
        else if ((*cp >= L'0') && (*cp <= L'9'))
            val = *cp - L'0' + 52;
        else if ((*cp == L'+') || (*cp == L'!'))
            val = 62;
        else if ((*cp == L'/') || (*cp == L'*'))
            val = 63;
        else if (*cp == L'=') {
            if (count == 2) {
                *ocp++ = block >> 4;
                ocount++;
            } else {
                *ocp++ = (block >> 10);
                *ocp++ = (block >> 2) & 0xff;
                ocount += 2;
            }
            break;
        } else
            val = 1;

        if (val >= 0) {
            block = (block << 6) | val;
            if (++count == 4) {
                *ocp++ = block >> 16;
                *ocp++ = (block >> 8) & 0xff;
                *ocp++ = block & 0xff;
                ocount += 3;
                count = 0;
            }
        }
    }
    return ocount;
}

/* Stupid little fake XML parser. */

static wchar_t *find_close(wchar_t * str)
{
    while ((*str != L'\0') && (*str != L'>')) {
        if (*str == L'"') {
            if ((str = wcschr(str + 1, L'"')) == NULL)
                return NULL;
        }
        str++;
    }

    if (*str == L'\0')
        return NULL;
    else
        return str + 1;
}



wchar_t *get_element(wchar_t * tag, wchar_t * str)
{
    int len = wcslen(tag);
    wchar_t *tmptag;
    wchar_t *start, *end;
    wchar_t *rval = NULL;
    wchar_t *tmp;

    if ((tmptag = malloc((len + 4) * sizeof(wchar_t))) == NULL) {
        lit_error(ERR_R, "Memory allocation failed in get_element()"); 
        return NULL;
    }
    swprintf(tmptag, L"<%s", tag);

    while (1) {
        if ((start = wcsstr(str, tmptag)) == NULL)
            goto exit;
        if (!iswalnum(start[len + 1]))
            break;
        str = start + len + 1;
    }

    swprintf(tmptag, L"</%s>", tag);
    end = wcsstr(str, tmptag);
    if (end == NULL) {
        goto exit;
    } else {
        wchar_t *realstart = find_close(start);
        if ((realstart == NULL) || (realstart > end)) 
            goto exit;
        tmp = malloc((end - realstart + 1) * sizeof(wchar_t));
        if (tmp == NULL) {
            lit_error(ERR_R, "Memory allocation #2 failed in get_element()"); 
            return NULL;
        }
        memcpy(tmp, realstart, (end - realstart) * sizeof(wchar_t));
        tmp[end - realstart] = L'\0';
        rval = tmp;
    }
exit:
    free(tmptag);
    return rval;
}

int drm5_handle_key(lit_file * plf)
{
    U8  * pData;
    int size, r;
    wchar_t  *ebits = NULL, * value = NULL;
    unsigned char * keyData;
    int keySize, status;
    DWORD extra;
    U8 * pValidation;

    status = lit_get_file(plf,license_string,&pData,&size);
    if (status) {
        lit_error(ERR_R, 
"Was unable to find an End-User-License (\"%s\") in the LIT file.", 
	license_string); 
        return status;
    }
    ebits = get_element(L"ENABLINGBITS", (wchar_t *)pData);
    if (ebits) value = get_element(L"VALUE", ebits);
    if (!value) {
        lit_error(ERR_R, 
"I couldn't find an encrypted key in the End User License file.");
        free(pData);
        return -1;
    }
    keySize = MS_Base64Decode(value, &keyData);
    extra = 0;
    do 
    {
        r = -1;
        __try
        {
            r = drm5hook_setkey(keyData, keySize,extra);
        }
        __except(EXCEPTION_EXECUTE_HANDLER)
        {
            lit_error(ERR_R,"Activation DLL doesn't like new method.");
            break;
        }
        if (r) {
            lit_error(ERR_R,"Failed to set a key for this title.");
            break;
        } else {
            status = lit_get_file(plf,"/DRMStorage/ValidationStream",
                &pValidation, &size);
            if ((status) || (!pValidation) || (!size)) {
                lit_error(ERR_R,"Unable to read validation stream.");
                free(pData);
                return -1;
            }
            if (strncmp(pValidation,"MSReader",8) != 0) { r = -1; }
            if (r == -1)
            {
                int i;
                /* Cleanse the bad sections */
                for (i = 0; i < plf->num_sections; i++)
                {
                    if ((plf->sections[i].data_pointer) && 
                        (plf->sections[i].data_pointer != (U8 *)-1))
                    {
                        free(plf->sections[i].data_pointer);
                        plf->sections[i].data_pointer = NULL;
                    }
                }
            }
            if (!r) { break; }
        }
        extra ++;
        if (extra > 128)
        {
            lit_error(ERR_R,"Unable to find a working key value!!");
            r = -1;
            break;
        }
    } while (r != 0);
    free(pData);
    free(pValidation);
    return r;
}

char * get_registry_entry(HKEY hkey, char * subkey, char * value, int pad)
{
    long    r;
    char    * buff;
    U32     len, type;
    HKEY    key;

    r = RegOpenKeyExA(hkey,subkey,0,KEY_READ,&key);
    if (r) return NULL;
    r = RegQueryValueExA(key,value,NULL,NULL,NULL,&len);
    if (r) return NULL;
    buff = (char *)malloc(len + pad);
    if (!buff) {
        lit_error(ERR_R, "Unable to get memory for registry entry.");
        return NULL;
    }
    r = RegQueryValueExA(key,value,NULL,&type,buff, &len);
    if (r) { free(buff); return NULL;}
    RegCloseKey(key);
    return buff;
}

int drm5_initialize_dll()
{
    HMODULE hDLL;
    int     r;
    char    * path;
    void    * pv;

    hDLL =  LoadLibrary("drm5hook.dll");
    if (!hDLL) {
        lit_error(ERR_WIN32,"Error loading DRM5HOOK.DLL!");
        return -1;
    } 
    pv = drm5hook_decrypt = 
        (p_drm5hook_decrypt)GetProcAddress(hDLL,"drm5hook_decrypt");
    if (pv)pv = drm5hook_init = 
        (p_drm5hook_init)GetProcAddress(hDLL,"drm5hook_init");
    if (pv)pv = drm5hook_install = 
        (p_drm5hook_install)GetProcAddress(hDLL,"drm5hook_install");
    if (pv)pv = drm5hook_setkey = 
        (p_drm5hook_setkey)GetProcAddress(hDLL,"drm5hook_setkey");
    if (!pv) {
        lit_error(0,"Error getting functions from DRM5HOOK.DLL!");
        return -1;
    }

    path = get_registry_entry(HKEY_LOCAL_MACHINE,
        "Software\\Microsoft\\EBook", "InstallDir", 20);
   
    if (path) {
        strcat(path+strlen(path),"MSREADER.EXE");
        r = drm5hook_install(path, (void *)&lit_error);
    }
    else 
        r = drm5hook_install("MSREADER.EXE", (void *)&lit_error);
    if (path) free(path);
    return r;
}

/*--[drm5_callback]------------------------------------------------------------
 |
 | This routine will handle interfacing to the main lit library.  
 |
 | If p is NULL, then initialize..
 | Otherwise, decrypt and hand back the data.
 | 
*/
int drm5_callback(void * pv, U8 * p, int size)
{
    int status = 0;
    if (!p) {
        status = drm5_initialize_dll();
        if (!status) status = drm5hook_init();
        if (!status) status = drm5_handle_key((lit_file *)pv);
    }
    else { status =  drm5hook_decrypt(p, size); };
    return status;
}
