/*--[drm5remt.c]---------------------------------------------------------------
 | Copyright (C) 2002, 2003 Dan A. Jackson 
 |
 | This file is part of the "clit" (Convert LIT) program. 
 |
 | Convert LIT is free software; you can redistribute it and/or modify
 | it under the terms of the GNU General Public License as published by
 | the Free Software Foundation; either version 2 of the License, or
 | (at your option) any later version.
 |
 | This program is distributed in the hope that it will be useful,
 | but WITHOUT ANY WARRANTY; without even the implied warranty of
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 | GNU General Public License for more details.
 |
 | You should have received a copy of the GNU General Public License
 | along with this program; if not, write to the Free Software
 | Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 | 
 | The GNU General Public License may also be available at the following
 | URL: http://www.gnu.org/licenses/gpl.html
 */
#include "drm5hook.h"
#include <stdlib.h>
#include <direct.h>


int server_message_loop(void);


typedef int (*voidfunc)(void);

typedef int (__stdcall *bbfunc)(void **);

typedef int (*initfunc)(void *,int, void *);

typedef int (*setrsafunc)(void *,int, unsigned char *, unsigned long int *);

typedef int (*setsymfunc)(void *,int, int, void *, unsigned long int *);

typedef int (*decfunc)(void *,unsigned long int,int,unsigned char *, unsigned long int *, unsigned char *);



typedef struct bbtable {

    void * f1;

    void * f2;

    initfunc init;

    void * f3;

    setrsafunc setrsa;

    void * f5;

    setsymfunc setsym;

    decfunc    decrypt;

    DWORD      test[64];
} bbtable;



extern int jumpgate(void);


int      paramcount = 0;

voidfunc trampoline_target = NULL;

unsigned long int ret_addr = 0;

void * bbobj;

bbtable * pTable;

unsigned int key_tag;

HMODULE  hDRM;

int   drm5_initialize();
int   drm5_setkey(BYTE * key, DWORD len, DWORD extra);
int   drm5_decrypt(BYTE * data, DWORD len);

void buffer_error(int what, char * fmt, ...) {
    int len;

    len = strlen(fmt);
    if (len > BUFFER_SIZE) len = BUFFER_SIZE-1;
    CopyMemory(buffer+16,fmt, len);
    *(buffer+16+(BUFFER_SIZE-1)) = '\0';
    *(DWORD *)(buffer+4) = len;
    *(DWORD *)(buffer) = RSP_ERROR;
}

char * get_registry_entry(HKEY hkey, char * subkey, char * value, int pad)

{

    long    r;

    char    * buff;

    DWORD   len, type;

    HKEY    key;



    r = RegOpenKeyExA(hkey,subkey,0,KEY_READ,&key);

    if (r) return NULL;

    r = RegQueryValueExA(key,value,NULL,NULL,NULL,&len);

    if (r) return NULL;

    buff = (char *)malloc(len + pad);

    if (!buff) {

        common_error_handler(0, 
            "Unable to get memory for registry entry %1!x! \"%1!s!\".",
                hkey, subkey);

        return NULL;

    }

    r = RegQueryValueExA(key,value,NULL,&type,buff, &len);

    if (r) { free(buff); return NULL;}

    RegCloseKey(key);

    return buff;

}


void drm5hook_self_initialize( char * szMap, char * szExec, char * szDone ) {
    int     finished;

    hMap = hExec = hDone = 0;
    finished = 0;
    buffer = NULL;
    do {
        hMap = OpenFileMapping(FILE_MAP_WRITE, FALSE, szMap); 
        if (!hMap) { 
            common_error_handler(TRUE, "Failed OpenFileMapping on %1!s!",szMap);
            break;
        }
        buffer = MapViewOfFile(hMap, FILE_MAP_WRITE, 0, 0, BUFFER_SIZE);
        if (!buffer) {
            common_error_handler(TRUE, "Failed MapViewOfFile on %1!s!", szMap);
            break;
        }
        hExec = OpenEvent(EVENT_ALL_ACCESS, FALSE, szExec); 
        if (!hExec) {
            common_error_handler(TRUE, "Failed OpenEvent on %1!s!",szExec);
            break;
        }
        hDone = OpenEvent(EVENT_ALL_ACCESS, FALSE, szDone); 
        if (!hDone) {
            common_error_handler(TRUE, "Failed OpenEvent on %1!s!",szDone);
            break;
        }
        finished++;
    } while(0);
    if (!finished) {
        common_destroy_global_state();
        ExitThread(-1);
    }
    finished = server_message_loop();
    common_destroy_global_state();
    return;
}

int server_message_loop(void)
{
    DWORD   retcode;
    DWORD   command, length;
    BYTE    * data;
    int     err;

    SetEvent(hDone);  
    while (1) {
        retcode = WaitForSingleObject(hExec, 30*1000); 
        if (retcode == WAIT_TIMEOUT) {
/*
            MessageBox(NULL,"Application Timeout!","DLL", MB_ICONERROR|MB_OK);
*/
            return 1;
        } else if (!retcode) {
            command = *(DWORD *)buffer;
            length = *(DWORD *)(buffer+4);
            data = (buffer+16);
            err = 0;
            error_handler = buffer_error; 

            switch (command) {
            case CMD_INIT:
                err = drm5_initialize();
                break;
            case CMD_DECRYPT:
                err = drm5_decrypt(data, length);
                break;
            case CMD_SETKEY:
                *(DWORD *)(buffer+4) = 0;
                err = drm5_setkey(data+4, length-4, *(DWORD *)data);
                break;
            default:
                ;
            }
            if ((err) && (*(DWORD *)buffer != RSP_ERROR))
            {
                common_error_handler(0,"CMD %1!d! ERR %2!d! \n",command, err);
            }
            error_handler = NULL;

            if (!err) {
                if (*(DWORD *)buffer != RSP_ERROR) *(DWORD *)buffer = RSP_OK;
            }

            SetEvent(hDone);
        }
    }
}



int  get_code_info(HMODULE image_base, DWORD   * pBase, DWORD   * pSize)

{

    PIMAGE_DOS_HEADER pmz;

    PIMAGE_NT_HEADERS ppe;

    PIMAGE_OPTIONAL_HEADER popthdr;



    pmz = (PIMAGE_DOS_HEADER)image_base;

    if (pmz->e_magic != IMAGE_DOS_SIGNATURE) return 0x10;



    ppe = (PIMAGE_NT_HEADERS)((BYTE *)image_base + pmz->e_lfanew);

    if (ppe->Signature != IMAGE_NT_SIGNATURE) return 0x20;

    if (ppe->FileHeader.SizeOfOptionalHeader == 0) return 0x30;

    popthdr = (PIMAGE_OPTIONAL_HEADER)&ppe->OptionalHeader;

    if (popthdr->Magic != IMAGE_NT_OPTIONAL_HDR_MAGIC) return 0x40;



    if (pBase) *pBase = popthdr->BaseOfCode;

    if (pSize) *pSize = popthdr->SizeOfCode;

    return 0;

}


int drm5_load_secure_repository(void)
{
    char * path;
    bbfunc  func;

    volatile initfunc hackinit;

    int     r, finished;

    path = hDRM = NULL;
    finished = 0; 
    do { 
        path = get_registry_entry(HKEY_CURRENT_USER,

            "Software\\Microsoft\\DRM","SecRepLocation",0);

        if (!path) {

            common_error_handler(FALSE, 
                "No DRM path in the registry - are you activated??");

            break;
        }
    
        hDRM = LoadLibraryExA(path, NULL, 0);

        if (!hDRM) {

            common_error_handler(TRUE, 
	            "Error loading DRM DLL: \"%1!s!\" !",path);

            break;
        }

  

        func = (bbfunc)GetProcAddress(hDRM,"DllGetBlackBox");

        if (!func) {

            common_error_handler(FALSE,
                "Unable to find DllGetBlackBox in \"%1!s!\".",path); 

            break;
        }

    
        free(path); 

        path = NULL;



        r = func(&bbobj);

        if ((r) || (!bbobj)) {

            common_error_handler(FALSE,
                "DllGetBlackBox() call failed! Status=%1!d!.",r);

            break;
        }
        
        pTable = (bbtable *)( *(bbtable **) bbobj );



        trampoline_target = (voidfunc)pTable->init;

        hackinit = (initfunc)jumpgate;

        paramcount = 3;

        r = hackinit(bbobj, GetCurrentProcessId(), NULL);

        if (r < 0) { 

            common_error_handler(FALSE,
                "Blackbox \"Initialize\" failed! (%1!d!)\n", r);

            break;
        }

        finished++;
    } while (0);
    if (!finished) {
        if (path) free(path);
        if (hDRM) FreeLibrary(hDRM);
        hDRM = path = NULL;
        return -1;
    }
    return 0;

}


LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch(msg) {
    case WM_CLOSE: DestroyWindow(hwnd); break;
    case WM_DESTROY: PostQuitMessage(0); break;
    default: return DefWindowProc(hwnd, msg, wParam, lParam);
    }
    return 0;
}

int drm5_spoof_checks()
{
    HINSTANCE hInstance;
    WNDCLASSEX wc;
    HWND hwnd;
    char * szClassName = "EBoOk_FrAMe";

    hInstance = GetModuleHandle(0);      
        wc.cbSize   = sizeof(WNDCLASSEX);
    wc.style   = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc  = WndProc;
    wc.cbClsExtra  = 0;
    wc.cbWndExtra  = 0;
    wc.hInstance  = hInstance;
    wc.hIcon   = LoadIcon(NULL, IDI_APPLICATION);
    wc.hIconSm   = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor   = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = szClassName;
    if(!RegisterClassEx(&wc))
    {
        MessageBox(NULL, "Window Registration Failed!", "Error!",
                   MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }
    hwnd = CreateWindowEx( 0, szClassName, "", 0, CW_USEDEFAULT, 
               CW_USEDEFAULT, 1, 1, NULL, NULL, hInstance, NULL);
    if(hwnd == NULL)
    {
        MessageBox(NULL, "Window Creation Failed!", "Error!",
                   MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }
    SetActiveWindow(hwnd); 

    hwnd = GetActiveWindow();
    if (!hwnd)
    {
        MessageBox(NULL," Failed to GetActiveWindow()","Bad...",
                    MB_OK);
    }
    (void)CreateMutexA(NULL, FALSE,"EBookLitPathMutex");

    return 0;
}

char * liblist[] = { "EBOOKFX.DLL","EBRICHED.DLL",

 "DMGR.DLL","MSLS31.DLL" };



int drm5_initialize() {
    char * path, *pathfile;

    int     r, finished;

    unsigned int i;
    HMODULE hReader, hLib;

    unsigned char * pTrash;

    DWORD     code_base, code_size;

    char    * old_directory;


    old_directory = path = NULL;

    finished = 0;
    do {
        path = get_registry_entry(HKEY_LOCAL_MACHINE,

            "Software\\Microsoft\\EBook", "InstallDir", 20);

        if (!path) {

            common_error_handler(FALSE,
                "No Registry key to locate MSREADER.EXE directory! \n");
            break;
        }
        pathfile = path+strlen(path);


        /* Fix search-path issue on Win9X */

        old_directory = _getcwd(NULL, 512);

        if (!old_directory) {

            common_error_handler(FALSE, 
                "Unable to get current working directory!");
        }

        if (old_directory) {
 _chdir(path);
 }



        hReader = GetModuleHandle(0); 

        for (i = 0; i < sizeof(liblist)/sizeof(liblist[0]); i++) {

            strcpy(pathfile, liblist[i]);

            hLib = LoadLibraryExA(path, NULL, DONT_RESOLVE_DLL_REFERENCES); 

            if (!hLib) {

                common_error_handler(TRUE,
                    "LoadLibrary Failed for \"%1!s!\".", liblist[i]);
                break;
            }

        }

        if (old_directory) { _chdir(old_directory); free(old_directory); }

        old_directory = NULL;

 

        r = get_code_info(hReader, &code_base, &code_size);

        if (r) { 

            common_error_handler(0,

"I couldn't understand the PE Header for MSREADER.EXE (Code: %1!d!).", 

                r);

            break;
        }    

        ret_addr = 0;

        pTrash = (unsigned char *)hReader + code_base;

        for (i = 5; i < code_size; i++) {

            if ( *(pTrash + i) == 0xC3) {

                if ( (*(pTrash + i - 5) == 0xFF) ||

                     (*(pTrash + i - 5) == 0xE8) ||

                    (*(pTrash + i - 3) == 0xFF) ||

                    (*(pTrash + i - 3) == 0xE8) ) {

                    ret_addr = (unsigned long int)(pTrash + i);

                    break;

                }

            }

        }

        free(path); path = NULL;
    
        if (!ret_addr) {

            common_error_handler(0,

                "I can't find a decent thunkable spot.  No crack for you!");

            break;
        }

        path = get_registry_entry(HKEY_LOCAL_MACHINE,

            "Software\\Microsoft\\EBook", "ClearTypeLocation", 0);

        if (!path) {

            common_error_handler(FALSE,
                "Unable to locate registry entry for ClearType DLL!"); 

            return -1;

        }

        hLib = LoadLibrary(path);

        if (!hLib) {

            common_error_handler(TRUE,"DLL load failed for \"%1!s!\" !",path);
            break;
        }
        free(path);
        path = NULL;
        finished++;
    }
    while (0);
    if (!finished) {
        if (path) free(path);

        if (old_directory) { _chdir(old_directory); free(old_directory); }

        path = old_directory = NULL;

        return -1;

    }


    r = drm5_spoof_checks();
    

    r = drm5_load_secure_repository();
    return r;
    
}


int drm5_setkey(BYTE * p, DWORD size, DWORD extra) {
    setsymfunc hacksetsym;

    int r, retval;



    retval = 0;

    trampoline_target = (voidfunc)pTable->setsym;

    hacksetsym = (setsymfunc)jumpgate;

    paramcount = 5;

    r = hacksetsym(bbobj, 0,size|(extra << 16), p, &key_tag);

    if (r < 0) {

        common_error_handler(FALSE,
            "Blackbox \"Set Symmetric Key\" failed! (%1!d!,%2!d!,%3!d!\n", 
                r, size, extra);

        return -1;

    }

    return 0;
}

int drm5_decrypt(BYTE * p, DWORD size) {

    decfunc    hackdec;

    int        result_size, r;



    result_size = size;

    trampoline_target = (voidfunc)pTable->decrypt;

    hackdec = (decfunc)jumpgate;

    paramcount = 6;

    r = hackdec(bbobj, key_tag,size,p,&result_size,p);

    if (r < 0) {

        common_error_handler(FALSE,
            "Blackbox \"Decryption\" failed! (%1!d!)\n", r);

        return -1;

    }

    return 0;

}


